use std::{collections::HashSet, path::Path};

use actix_web::{http::StatusCode, web::Query, HttpRequest, HttpResponse, HttpResponseBuilder};
use rand::{rngs::SmallRng, seq::SliceRandom, Rng, SeedableRng};
use serde::Deserialize;
use tracing::{debug, error, info, log::warn};

use crate::{
    statsapi::{get_player_stats, Stats},
    steamapi::get_player,
    template::{self, GLOBAL_CONTEXT, TERA},
    CONFIGURATION,
};

lazy_static! {
    pub static ref MAPS: HashSet<&'static str> = {
        info!("Getting map images…");
        let mut maps: HashSet<&'static str> = HashSet::new();
        let mut search_dir = Path::new(&CONFIGURATION.web.template_dir).to_path_buf();
        search_dir.push("processed");
        let dir_iter = match std::fs::read_dir(search_dir) {
            Ok(v) => v,
            Err(e) => {
                warn!("Unable to read background maps dir '{}'", e);
                return maps;
            }
        };
        for file in dir_iter {
            match file {
                Ok(f) => {
                    if f.file_name().to_string_lossy().ends_with("_processed.png") {
                        let fname = f.file_name();
                        let lossy = fname.to_string_lossy();
                        let mapname = lossy
                            .strip_suffix("-1_processed.png")
                            .or_else(|| lossy.strip_suffix("-2_processed.png"));
                        if mapname.is_none() {
                            continue;
                        }
                        maps.insert(Box::leak(
                            mapname
                                .unwrap_or("never_reached")
                                .to_string()
                                .into_boxed_str(),
                        ));
                    }
                }
                Err(e) => {
                    warn!("Unable to read file: '{}'", e)
                }
            }
        }
        info!("Done!");
        debug!("Maps Found: {:?}", maps);
        maps
    };
}

#[tracing::instrument(level = "info")]
pub async fn loadingurl(req: HttpRequest) -> HttpResponse {
    let mut context = GLOBAL_CONTEXT.clone();
    // Get Queries
    let query_str = req.query_string();
    let request_data = match Query::<LoadingUrlRequest>::from_query(query_str) {
        Ok(v) => v.into_inner(),
        Err(e) => {
            error!("Couldn't extract loadingurl query information: {e}");
            LoadingUrlRequest::default()
        }
    };
    // Insert them
    let mapname = request_data.mapname.unwrap_or_default();
    context.insert("map_exists", &MAPS.contains(mapname.as_str()));
    context.insert("mapname", &mapname);
    // Get SteamId Player
    let player = get_player(request_data.steamid.unwrap_or_default())
        .await
        .map_err(|e| {
            // Inspect err is unstable rn, so map it is.
            error!("Error getting player: {e}");
            e
        })
        .unwrap_or_default();
    context.insert("player", &player);
    let stats = get_player_stats(&player)
        .await
        .map_err(|e| {
            // Inspect err is unstable rn, so map it is.
            error!("Error getting stats: {e}");
            e
        })
        .unwrap_or_default();
    context.insert("stats", &stats);
    // Flare
    let mut rng = SmallRng::from_entropy();
    let flares = random_flare(&mut rng);
    let fun_fact = fun_fact(&mut rng, &stats);
    context.insert("greeting", flares.0);
    context.insert("verb", flares.1);
    context.insert("fun_fact", &fun_fact);
    // Finish up.
    let bdy = match TERA.render("loadingurl.html", &context) {
        Ok(v) => v,
        Err(e) => format!("{:#?}", e), // Errors just get displayed.
    };
    HttpResponseBuilder::new(StatusCode::OK).body(bdy)
}

#[tracing::instrument(level = "info")]
pub async fn precache(req: HttpRequest) -> HttpResponse {
    // Get Queries
    let request_data = match Query::<LoadingUrlRequest>::from_query(req.query_string()) {
        Ok(v) => v.into_inner(),
        Err(e) => {
            error!("Couldn't extract loadingurl query information: {e}");
            return HttpResponseBuilder::new(StatusCode::BAD_REQUEST).body(e.to_string());
        }
    };
    // Cache SteamIdPlayer
    match get_player(request_data.steamid.unwrap_or_default()).await {
        Ok(_) => HttpResponseBuilder::new(StatusCode::OK).finish(),
        Err(e) => {
            HttpResponseBuilder::new(StatusCode::INTERNAL_SERVER_ERROR).body(format!("{:?}", e.source()))
        }
    }
}

#[derive(Debug, Deserialize)]
pub struct LoadingUrlRequest {
    pub mapname: Option<String>,
    pub steamid: Option<u64>,
}

impl Default for LoadingUrlRequest {
    fn default() -> Self {
        Self {
            mapname: None,
            steamid: Default::default(),
        }
    }
}

fn random_flare(rng: &mut SmallRng) -> (&'static str, &'static str) {
    let greeting = *template::GREETINGS.choose(rng).unwrap_or(&"Default");
    let verb = *template::VERBS.choose(rng).unwrap_or(&"Default");
    (greeting, verb)
}

fn fun_fact(rng: &mut SmallRng, stats: &Stats) -> String {
    match rng.gen_range(0..3) {
        0 => format!("You have died {} times.", stats.deaths),
        1 => format!("You have fired {} bullets.", stats.shots),
        2 => format!("You have suicided {} times.", stats.suicides),
        _ => "There is no such thing as a fun fact.".into(),
    }
}
