pub const AUTHORS: &str = env!("CARGO_PKG_AUTHORS");
pub const VERSION: &str = env!("CARGO_PKG_VERSION");
pub const BUILDTD: &str = env!("BUILD_TIMEDATE");
pub const REVISION: &str = env!("VCS_COMMIT_HASH");
pub const REVDATE: &str = env!("VCS_COMMIT_DATE");
pub const PROFILE: &str = env!("COMPILE_PROFILE");

mod configuration;
mod loadingurl;
mod statsapi;
mod steamapi;
mod template;

use crate::configuration::Conf;

#[macro_use]
extern crate lazy_static;

use actix_web::{middleware, rt, web, App, HttpServer};
use eyre::Context;

use tracing::{event, info};

// This is declared at the crate root instead of inside crate::Configuration
// because it's easier to reference that way.
lazy_static! {
    pub static ref CONFIGURATION: Conf =
        configuration::get_config_init_logging().expect("Unable to load configuration!");
}

fn main() -> Result<(), eyre::Error> {
    println!("╔════════════╤═════════ LOADING URL ════════════{PROFILE:═>7}═══╗");
    println!("║ Authors    │ {0: <43}║", AUTHORS);
    println!("║ Version    │ {0: <43}║", VERSION);
    println!("║ Build Date │ {0: <43}║", BUILDTD);
    println!("║ Revision   │ {0: <43}║", REVISION);
    println!("║ Rev Date   │ {0: <43}║", REVDATE);
    println!("╚════════════╧════════════════════════════════════INIT════╝\n");

    // Some of these could panic, so do that before we start the runtime.
    lazy_static::initialize(&CONFIGURATION);
    lazy_static::initialize(&template::TERA);
    lazy_static::initialize(&loadingurl::MAPS);
    lazy_static::initialize(&template::GREETINGS);
    lazy_static::initialize(&template::VERBS);
    lazy_static::initialize(&steamapi::STEAMID_CACHE);
    lazy_static::initialize(&statsapi::STATS_CACHE);

    event!(tracing::Level::INFO, "Initialized\n");
    real_main().wrap_err(format!("WHILE RUNNING"))
}

fn real_main() -> Result<(), eyre::Error> {
    let ipaddr = CONFIGURATION.core.socketaddr;
    let system = rt::System::new();
    let applcl = || {
        let app = App::new()
            .wrap(
                middleware::Logger::new(r#"[%a | %{r}a] %s "%r" %b '%{User-Agent}i' %Dms"#)
                    .log_target("HTTP >"),
            )
            .service(web::resource("/loadingurl/").route(web::get().to(loadingurl::loadingurl)))
            .service(
                web::resource("/loadingurl/precache/").route(web::get().to(loadingurl::precache)),
            );
        match &CONFIGURATION.web.serve_assets {
            Some(d) => {
                info!("Serving assets.");
                app.service(actix_files::Files::new("/loadingurl/", d))
            }
            None => app,
        }
    };
    let server = HttpServer::new(applcl)
        .bind(ipaddr)
        .wrap_err(format!("Trying to bind to: {}", ipaddr))?;
    info!("Actix bound to: {}", ipaddr);
    let worker_count = &CONFIGURATION
        .core
        .threads
        .unwrap_or(num_cpus::get_physical());
    let running = server.workers(*worker_count).run();
    // RUNTIME
    system.block_on(running).wrap_err("Executing Actix future")
}
