use std::{fs::File, io::Read, path::PathBuf};

use serde::{de, Deserialize, Serialize};
use tera::{Context, Tera};
use tracing::{error, event, Level};

use crate::{CONFIGURATION, REVDATE, REVISION};

lazy_static! {
    pub static ref TERA: Tera = {
        let _span = tracing::span!(Level::INFO, "Tera").entered();
        event!(Level::INFO, "Compiling…");
        // This is reletive to the CLI current dir 🡫
        let mut tera = match Tera::new(format!("{}/*.html", CONFIGURATION.web.template_dir).as_str())
            {
            Ok(t) => t,
            Err(e) => {
                panic!("Parsing error(s): {}", e);
            }
            };
        tera.autoescape_on(vec![".html"]);
        event!(Level::INFO, "Compiled.");
        tera
    };
}

lazy_static! {
    /// Contains a set of values reliably accessable from any render context
    ///
    /// Values:
    /// - `GLOBAL_ASSET_URL`
    /// - `GLOBAL_REVISION`
    /// - `GLOBAL_REVTD`
    /// - `GLOBAL_TICKER`
    /// - `GLOBAL_PAGEVERSION`
    pub static ref GLOBAL_CONTEXT: Context = {
        let mut ctx = Context::new();
        ctx.insert("GLOBAL_ASSET_URL", &CONFIGURATION.web.asset_url);
        ctx.insert("GLOBAL_REVISION", REVISION);
        ctx.insert(
            "GLOBAL_REVTD",
            REVDATE,
        );
        ctx.insert("GLOBAL_TICKER", *TICKER);
        ctx.insert("GLOBAL_PAGEVERSION", &(*PAGEVERSION));
        ctx
    };
}
fn file_or_default(path: &PathBuf) -> String {
    let mut string = String::new();
    match File::open(&path) {
        Ok(mut f) => {
            f.read_to_string(&mut string).unwrap_or_else(|_| 0);
        }
        Err(e) => {
            error!("Unable to read '{}' - '{e}'", &path.to_string_lossy());
        }
    };
    string
}

// Bunch of strings we're loading at startup and leaking to be static.
lazy_static! {
    pub static ref GREETINGS: &'static Vec<&'static str> = {
        let mut path = PathBuf::from(&CONFIGURATION.web.template_dir);
        path.push("greetings.txt");
        Box::leak(Box::new(
            Box::leak(file_or_default(&path).into_boxed_str())
                .lines()
                .collect::<Vec<&str>>(),
        ))
    };
}

lazy_static! {
    pub static ref VERBS: &'static Vec<&'static str> = {
        let mut path = PathBuf::from(&CONFIGURATION.web.template_dir);
        path.push("verbs.txt");
        Box::leak(Box::new(
            Box::leak(file_or_default(&path).into_boxed_str())
                .lines()
                .collect::<Vec<&str>>(),
        ))
    };
}

lazy_static! {
    pub static ref TICKER: &'static Vec<&'static str> = {
        let mut path = PathBuf::from(&CONFIGURATION.web.template_dir);
        path.push("ticker.txt");
        Box::leak(Box::new(
            Box::leak(file_or_default(&path).into_boxed_str())
                .lines()
                .collect::<Vec<&str>>(),
        ))
    };
}

lazy_static! {
    pub static ref PAGEVERSION: VersionInfo = {
        let mut path = PathBuf::from(&CONFIGURATION.web.template_dir);
        path.push("versioninfo.json");
        let vstring = file_or_default(&path);
        let mut version: VersionInfo = serde_json::from_str(&vstring)
            .unwrap_or_else(|e| {
                    error!("Unable to deserialize versioninfo.json: '{e}'");
                    VersionInfo {
                        version: "ffffffffffff".into(),
                        modified: "00000000".into(),
                    }
                }
            );
        // Nasty formatting quickness, this can panic.
        let base = format!("{:0>8}", version.modified);
        let year = &base[..4];
        let month = &base[4..6];
        let day = &base[6..8];
        version.modified = format!("{year}-{month}-{day}");
        version
    };
}

#[derive(Debug, Serialize, Deserialize)]
pub struct VersionInfo {
    pub version: String,
    #[serde(deserialize_with = "deserialize_date")]
    pub modified: String,
}

// We need to hoist a u64 into a String during deserialization.
fn deserialize_date<'de, D>(deserializer: D) -> Result<String, D::Error>
where
    D: de::Deserializer<'de>,
{
    let num: u64 = de::Deserialize::deserialize(deserializer)?;
    Ok(format!("{:0>8}", num))
}
